# (C) Copyright 2020- ECMWF.  #
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging

LOG = logging.getLogger(__name__)

PRIVATE_ATTRIBUTES = {"observer": lambda: None}


class MetaBase(type):
    def __call__(cls, *args, **kwargs):
        obj = cls.__new__(cls, *args, **kwargs)
        args, kwargs = cls.patch(obj, *args, **kwargs)
        obj.__init__(*args, **kwargs)
        return obj

    def patch(cls, obj, *args, **kwargs):
        private_attributes = {}
        private_attributes.update(PRIVATE_ATTRIBUTES)
        private_attributes.update(kwargs.pop("_PRIVATE_ATTRIBUTES", {}))

        for k, v in private_attributes.items():
            setattr(obj, k, kwargs.pop(k, v))

        return args, kwargs


class Base(metaclass=MetaBase):

    # Convertors
    def to_numpy(self, **kwargs):
        self._not_implemented()

    def to_xarray(self, **kwargs):
        self._not_implemented()

    def to_pandas(self, **kwargs):
        self._not_implemented()

    def to_tfdataset(self, **kwargs):
        self._not_implemented()

    def to_metview(self, **kwargs):
        self._not_implemented()

    # Change class
    def mutate(self):
        return self

    @classmethod
    def merge(cls, *args, **kwargs):
        return None

    # Used when plotting
    def plot_map(self, backend):
        self._not_implemented()

    def plot_graph(self, backend):
        self._not_implemented()

    def field_metadata(self):
        self._not_implemented()

    # I/O
    def save(self, path):
        self._not_implemented()

    def write(self, f):
        self._not_implemented()

    # Used by normalisers
    def to_datetime(self):
        self._not_implemented()

    def to_datetime_list(self):
        self._not_implemented()

    def to_bounding_box(self):
        self._not_implemented()

    # For machine learning
    def statistics(self):
        self._not_implemented()

    #
    def _not_implemented(self):
        import inspect

        func = inspect.stack()[1][3]
        module = self.__class__.__module__
        name = self.__class__.__name__

        extra = ""
        if hasattr(self, "path"):
            extra = self.path
        raise NotImplementedError(f"{module}.{name}.{func}({extra})")



# (C) Copyright 2021 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


import json
import os
import sqlite3

import requests
from multiurl import robust

from climetlab.core.caching import cache_file
from climetlab.utils import tqdm


def get_iterator_and_size(url):
    if os.path.exists(url):
        iterator = open(url).readlines()
        size = os.path.getsize(url)
        return iterator, size

    r = robust(requests.get)(url, stream=True)
    r.raise_for_status()
    try:
        size = int(r.headers.get("Content-Length"))
    except Exception:
        size = None
    iterator = r.iter_lines()
    return iterator, size


def create_table(target, names):
    if os.path.exists(target):
        os.unlink(target)
    connection = sqlite3.connect(target)

    sql_names = [f"i_{n}" for n in names]
    others = ",".join([f"{n} TEXT" for n in sql_names])

    create_statement = f"""CREATE TABLE entries (
        path    TEXT,
        offset  INTEGER,
        length  INTEGER,
        {others}
        );"""
    connection.execute(create_statement)

    commas = ",".join(["?" for _ in names])
    insert_statement = f"""INSERT INTO entries(
                                       path, offset, length, {','.join(sql_names)})
                                       VALUES(?,?,?,{commas});"""
    return connection, insert_statement, sql_names


class Database:
    def lookup(self, request):
        raise NotImplementedError("")


class SqlDatabase(Database):
    VERSION = 2

    def __init__(
        self,
        url,
        create_index=False,  # index is disabled by default because it is long to create.
    ):
        self._connection = None
        self.url = url
        self.create_index = create_index

    @property
    def connection(self):
        if self._connection is None:
            path = cache_file(
                "index",
                self.to_sql_target,
                self.url,
                hash_extra=self.VERSION,
                extension=".db",
            )
            self._connection = sqlite3.connect(path)
        return self._connection

    def to_sql_target(self, target, url):
        iterator, size = get_iterator_and_size(url)

        count = 0
        names = None
        connection = None
        insert_statement = None
        pbar = tqdm(
            iterator,
            desc="Downloading index",
            total=size,
            unit_scale=True,
            unit="B",
            leave=False,
            disable=False,
            unit_divisor=1024,
        )
        for line in pbar:
            entry = json.loads(line)
            if connection is None:
                # this is done only for the first entry only
                names = [n for n in sorted(entry.keys()) if not n.startswith("_")]
                connection, insert_statement, sql_names = create_table(target, names)

            assert names is not None, names
            assert connection is not None, connection
            assert insert_statement is not None
            assert len(sql_names) == len(names), (sql_names, names)

            # additional check is disabled
            # _names = [n for n in sorted(entry.keys()) if not n.startswith("_")]
            # assert _names == names, (names, _names)

            values = [entry.get("_path", None), entry["_offset"], entry["_length"]] + [
                entry[n] for n in names
            ]
            connection.execute(insert_statement, tuple(values))

            count += 1
            pbar.update(len(line) + 1)

        if self.create_index:
            # connection.execute(f"CREATE INDEX path_index ON entries (path);")
            for n in sql_names:
                connection.execute(f"CREATE INDEX {n}_index ON entries ({n});")

        connection.execute("COMMIT;")
        connection.close()

    def lookup(self, request):
        conditions = []
        for k, b in request.items():
            if isinstance(b, (list, tuple)):
                if len(b) == 1:
                    conditions.append(f"i_{k}='{b[0]}'")
                    continue
                w = ",".join([f"'{x}'" for x in b])
                conditions.append(f"i_{k} IN ({w})")
            else:
                conditions.append(f"i_{k}='{b}'")

        statement = f"SELECT path,offset,length FROM entries WHERE {' AND '.join(conditions)} ORDER BY offset;"

        parts = []
        for path, offset, length in self.connection.execute(statement):
            parts.append((path, (offset, length)))
        return parts



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

from climetlab import load_source
from climetlab.core.statistics import record_statistics

from .multi import MultiSource


class IndexedUrls(MultiSource):
    def __init__(
        self,
        index,
        request,
        *args,
        filter=None,
        merger=None,
        force=None,
        **kwargs,
    ):

        urls_parts = index.lookup_request(request)
        record_statistics(
            "indexed-urls",
            request=str(request),
        )

        sources = []
        for url, parts in urls_parts:
            source = load_source(
                "url",
                url=url,
                parts=parts,
                filter=filter,
                merger=merger,
                force=force,
                # Load lazily so we can do parallel downloads
                # lazily=True,
                **kwargs,
            )
            sources.append(source)
        if not sources:
            raise ValueError("Empty request: no match.")

        super().__init__(sources, filter=filter, merger=merger)


source = IndexedUrls



# (C) Copyright 2021 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


from .benchmarks.indexed_url import benchmark as benchmark_indexed_url
from .tools import experimental, parse_args


class BenchmarkCmd:
    @parse_args(
        indexedurl=dict(
            action="store_true",
            help="Benchmark on using indexed URL (byte-range) and various servers.",
        ),
        all=dict(action="store_true", help="Run all benchmarks."),
    )
    @experimental
    def do_benchmark(self, args):
        """Run predefined benchmarks, for CliMetLab development purposes."""
        if args.all or args.indexedurl:
            print("Starting benchmark.")
            benchmark_indexed_url()



# (C) Copyright 2021 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import json

from climetlab.utils import download_and_cache

URL = "https://apps.ecmwf.int/codes/grib/json/"


def grib_database():
    path = download_and_cache(URL)
    with open(path) as f:
        entries = json.load(f)

    for entry in entries["parameters"]:
        yield entry


if __name__ == "__main__":

    for n in grib_database():
        print(n)



# (C) Copyright 2021 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#
import inspect
import logging
import os
import re
import threading
from functools import wraps

from climetlab.utils import load_json_or_yaml
from climetlab.utils.availability import Availability

LOG = logging.getLogger(__name__)


def dict_args(func):
    @wraps(func)
    def wrapped(*args, **kwargs):
        m = []
        p = {}
        for q in args:
            if isinstance(q, dict):
                p.update(q)
            else:
                m.append(q)
        p.update(kwargs)
        return func(*m, **p)

    return wrapped


LOCK = threading.RLock()


def locked(func):
    @wraps(func)
    def wrapped(*args, **kwargs):
        with LOCK:
            return func(*args, **kwargs)

    return wrapped


class Decorator:

    is_availability = False

    def __call__(self, func):
        from climetlab.arguments import InputManager

        if not callable(func):
            manager = InputManager(decorators=[self])
            return manager.apply_to_value(func)

        decorators = [self]

        def unwrap(f):
            if not hasattr(f, "_climetlab_decorators"):
                return f
            return unwrap(f.__wrapped__)

        unwrapped = unwrap(func)

        if hasattr(func, "_climetlab_decorators"):
            decorators = decorators + func._climetlab_decorators

        manager = InputManager(decorators=decorators)

        @wraps(unwrapped)
        def newfunc(*args, **kwargs):
            args, kwargs = manager.apply_to_arg_kwargs(args, kwargs, func=unwrapped)
            return unwrapped(*args, **kwargs)

        newfunc._climetlab_decorators = decorators

        return newfunc


OPTIONS = {
    "date": ("format",),
    "date-list": ("format",),
    "bounding-box": ("format",),
    "bbox": ("format",),
    "variable": ("convention",),
    "variable-list": ("convention",),
}


class normalize(Decorator):
    def __init__(
        self,
        name,
        values=None,
        **kwargs,
    ):
        assert name is None or isinstance(name, str)
        self.name = name

        if isinstance(values, str):
            assert (
                kwargs.get("type") is None
            ), f"Cannot mix values={values} and type={kwargs.get('type')}"
            if "(" in values:
                m = re.match(r"(.+)\((.+)\)", values)
                type = m.group(1)
                args = m.group(2).split(",")
            else:
                type = values
                args = []

            # len(args) <= len(options)
            if args:
                for name, value in zip(OPTIONS[type], args):
                    kwargs[name] = value
            kwargs["type"] = type
        else:
            kwargs["values"] = values

        if "aliases" in kwargs and isinstance(kwargs["aliases"], str):
            _, ext = os.path.splitext(kwargs["aliases"])
            if ext in (".json", ".yaml", ".yml"):
                path = kwargs["aliases"]
                if not os.path.isabs(path):
                    caller = os.path.dirname(inspect.stack()[1].filename)
                    path = os.path.join(caller, path)
                kwargs["aliases"] = load_json_or_yaml(path)

        self.kwargs = kwargs

    def register(self, manager):
        manager.register_normalize(self)


class availability(Decorator):
    is_availability = True

    def __init__(self, availability, **kwargs):
        if isinstance(availability, str):
            if not os.path.isabs(availability):
                caller = os.path.dirname(inspect.stack()[1].filename)
                availability = os.path.join(caller, availability)

        self.availability = Availability(availability, **kwargs)

    def register(self, manager):
        manager.register_availability(self)



# (C) Copyright 2021 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

from collections import defaultdict

from climetlab.utils.patterns import Pattern

from .backends import IndexBackend, JsonIndexBackend


class Index:
    def __init__(self, backend=None) -> None:
        if backend is None:
            backend = JsonIndexBackend
        assert issubclass(backend, IndexBackend), backend
        self._backend_constructor = backend


class GlobalIndex(Index):
    def __init__(self, index_location, baseurl, backend=None) -> None:
        """The GloblaIndex has one index managing multiple urls/files.
        This unique index is found at "index_location"
        The path of each file is written in the index as a relative path.
        It is relative to a base url: "baseurl".
        """

        super().__init__(backend=backend)
        self.baseurl = baseurl
        # if is url index_location
        #   download index_location
        self.backend = self._backend_constructor(index_location)

    def get_backend(self, url=None):
        return self.backend

    def lookup_request(self, request):
        dic = defaultdict(list)

        # group parts by url
        for path, parts in self.backend.lookup(request):
            url = f"{self.baseurl}/{path}"
            dic[url].append(parts)

        # and sort
        dic = {url: sorted(parts) for url, parts in dic.items()}

        urls_parts = [(k, v) for k, v in dic.items()]

        return urls_parts


class PerUrlIndex(Index):
    # index2 : 1 index per url.
    #        2a: index location = url location + '.jsonl'
    #        2b: index location = url location - '.grib' + '.jsonl'
    # index = PerUrlIndex(pattern, backend=JsonIndexBackend)
    def __init__(
        self,
        pattern,
        backend=None,
        substitute_extension=False,
        index_extension=".index",
    ) -> None:
        """The PerUrlIndex uses one index for each urls/files that it manages.

        The urls are built from the pattern and the request.
        For each url, the location of the index is build (then downloaded and cached)
        using the url and changing its extension into ".index".

        pattern: pattern to build the list of url.
        index_extension (".index"): extension for the index url.
        substitute_extension (False): if set to True, substitute the index_extension
        to build the index url instead of appending.
        lookup_request(): get the urls and parts for a given request.
        """
        super().__init__(backend=backend)
        self.pattern = pattern
        self.substitute_extension = substitute_extension
        self.index_extension = index_extension
        self.backends = {}

    def _build_index_file(self, url):
        if callable(self.substitute_extension):
            return self.substitute_extension(url)
        if self.substitute_extension:
            url = url.rsplit(".", 1)[0]
        return url + self.index_extension

    def get_backend(self, url):
        if url in self.backends:
            return self.backends[url]
        assert isinstance(url, str), url

        index_url = self._build_index_file(url)
        backend = self._backend_constructor(index_url)

        self.backends[url] = backend
        return self.backends[url]

    def lookup_request(self, request):
        dic = defaultdict(list)

        pattern = Pattern(self.pattern, ignore_missing_keys=True)
        urls = pattern.substitute(**request)
        if not isinstance(urls, list):
            urls = [urls]

        request = dict(**request)  # deepcopy to avoid changing the user's request
        for used in pattern.names:
            # consume arguments used by Pattern to build the urls
            # This is to avoid keeping them on the request
            request.pop(used)

        # group parts by url
        for url in urls:
            backend = self.get_backend(url)
            for _, parts in backend.lookup(request):
                dic[url].append(parts)

        # and sort
        dic = {url: sorted(parts) for url, parts in dic.items()}

        urls_parts = [(k, v) for k, v in dic.items()]

        return urls_parts

    def __repr__(self) -> str:
        return f"PerUrlIndex(pattern={self.pattern})"



# (C) Copyright 2021 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import json
import os

import climetlab as cml

from .tools import parse_args


def _index_grib_file(path, path_name=None):
    import eccodes

    with open(path, "rb") as f:

        h = eccodes.codes_grib_new_from_file(f)

        while h:
            try:
                field = dict()

                if isinstance(path_name, str):
                    field["_path"] = path_name
                elif path_name is False:
                    pass
                elif path_name is None:
                    field["_path"] = path
                else:
                    raise ValueError(f"Value of path_name cannot be '{path_name}.'")

                i = eccodes.codes_keys_iterator_new(h, "mars")
                try:
                    while eccodes.codes_keys_iterator_next(i):
                        name = eccodes.codes_keys_iterator_get_name(i)
                        value = eccodes.codes_get_string(h, name)
                        field[name] = value

                finally:
                    eccodes.codes_keys_iterator_delete(i)

                field["_offset"] = eccodes.codes_get_long(h, "offset")
                field["_length"] = eccodes.codes_get_long(h, "totalLength")

                field["_param_id"] = eccodes.codes_get_string(h, "paramId")
                field["param"] = eccodes.codes_get_string(h, "shortName")

                yield field

            finally:
                eccodes.codes_release(h)

            h = eccodes.codes_grib_new_from_file(f)


def _index_url(path_name, url):
    path = cml.load_source("url", url).path
    # TODO: or use download_and_cache?
    # path = download_and_cache(url)
    yield from _index_grib_file(path, path_name=path_name)


def _index_path(path):
    if os.path.isdir(path):
        for root, _, files in os.walk(path):
            for name in files:
                yield from _index_grib_file(os.path.join(root, name))
    else:
        yield from _index_grib_file(path)


class GribCmd:
    @parse_args(
        paths_or_urls=dict(
            metavar="PATH_OR_URL",
            type=str,
            nargs="+",
            help="list of files or directories or urls to index",
        ),
        # json=dict(action="store_true", help="produce a JSON output"),
        baseurl=dict(
            metavar="BASEURL",
            type=str,
            help="Base url to use as a prefix to happen on each PATHS_OR_URLS to build urls.",
        ),
    )
    def do_index_gribs(self, args):
        """Create index files for grib files.
        If the option --baseurl is provided, create an index for multiple gribs.
        See https://climetlab.readthedocs.io/contributing/grib.html for details.
        """
        for path_or_url in args.paths_or_urls:
            if args.baseurl:
                entries = _index_url(
                    url=f"{args.baseurl}/{path_or_url}", path_name=path_or_url
                )
            elif os.path.exists(path_or_url):
                entries = _index_path(path_or_url)
            elif path_or_url.startswith("https://"):
                entries = _index_url(url=path_or_url, path_name=False)
            else:
                raise ValueError(f'Cannot find "{path_or_url}" to index it.')

            for e in entries:
                print(json.dumps(e))



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#
import keras
import tensorflow as tf

# https://github.com/tensorflow/docs/blob/304a7e4a90e53751ac59cab46667d78e81736fa3/site/en/guide/data_performance_analysis.md


def make_labels_hash_table(
    labels,
    name=None,
    key_dtype=None,
    dtype=tf.int64,
):

    if key_dtype is None:
        if all(isinstance(label, str) for label in labels):
            key_dtype = tf.string

        if all(isinstance(label, int) for label in labels):
            key_dtype = tf.int64

        if all(isinstance(label, float) for label in labels):
            key_dtype = tf.float64

    mapping = dict()
    for label in labels:
        mapping[len(mapping)] = label

    return mapping, tf.lookup.StaticHashTable(
        initializer=tf.lookup.KeyValueTensorInitializer(
            keys=tf.constant(labels, dtype=key_dtype),
            values=tf.constant(list(range(len(mapping))), dtype=dtype),
        ),
        default_value=tf.constant(-1, dtype=dtype),
        name=name,
    )


def make_label_one_hot(table, name=None, axis=-1):
    def wrapped(x):
        return tf.one_hot(
            table.lookup(x),
            int(table.size()),
            name=name,
            axis=axis,
        )

    return wrapped


class PeriodicConv2D(keras.layers.Conv2D):
    def call(self, inputs):
        w, h = self.kernel_size
        inputs = tf.concat([inputs, inputs[:, :, :w, :]], axis=2)
        inputs = tf.pad(
            inputs,
            [[0, 0], [h // 22, h // 2], [0, 0], [0, 0]],
            constant_values=0,
        )
        return super().call(inputs)



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import copy
import json
import logging
import math
import warnings

from climetlab.core.caching import auxiliary_cache_file
from climetlab.profiling import call_counter
from climetlab.sources import Source
from climetlab.utils.bbox import BoundingBox

from .codes import CodesReader, GribField, GribIndex

LOG = logging.getLogger(__name__)


def missing_is_none(x):
    return None if x == 2147483647 else x


def mix_kwargs(
    user,
    default,
    forced={},
    logging_owner="",
    logging_main_key="",
):
    kwargs = copy.deepcopy(default)

    for k, v in user.items():
        if k in forced and v != forced[k]:
            LOG.warning(
                (
                    f"In {logging_owner} {logging_main_key},"
                    f"ignoring attempt to override {k}={forced[k]} with {k}={v}."
                )
            )
            continue

        if k in default and v != default[k]:
            LOG.warning(
                (
                    f"In {logging_owner} {logging_main_key}, overriding the default value "
                    f"({k}={default[k]}) with {k}={v} is not recommended."
                )
            )

        kwargs[k] = v

    kwargs.update(forced)

    return kwargs


class FieldAdapter:
    def __init__(self, field, ignore_keys):
        self.field = field
        self.ignore_keys = ignore_keys

    def __getitem__(self, name):
        if name in self.ignore_keys:
            return None
        return self.field[name]


class FieldsetAdapter:
    def __init__(self, fieldset, ignore_keys):
        self.fieldset = fieldset
        self.ignore_keys = ignore_keys

    def __getitem__(self, n):
        return FieldAdapter(self.fieldset[n], self.ignore_keys)

    def __len__(self):
        return len(self.fieldset)


class FieldSet(Source):
    def __init__(self, *, paths=None):
        self._statistics = None
        self.readers = {}
        self.fields = []
        if paths is not None:
            if not isinstance(paths, (list, tuple)):
                paths = [paths]
            for path in paths:
                index = GribIndex(path)
                for offset, length in zip(index.offsets, index.lengths):
                    self.fields.append((path, offset, length))

    def reader(self, path):
        if path not in self.readers:
            self.readers[path] = CodesReader(path)
        return self.readers[path]

    def __getitem__(self, n):
        path, offset, length = self.fields[n]
        return GribField(self.reader(path), offset, length)

    def __len__(self):
        return len(self.fields)

    @property
    def first(self):
        return self[0]

    def to_tfdataset(
        self, split=None, shuffle=None, normalize=None, batch_size=0, **kwargs
    ):
        # assert "label" in kwargs
        if "offset" in kwargs:
            return self._to_tfdataset_offset(**kwargs)
        if "label" in kwargs:
            return self._to_tfdataset_supervised(**kwargs)
        else:
            return self._to_tfdataset_unsupervised(**kwargs)

    def _to_tfdataset_offset(self, offset, **kwargs):

        # μ = self.statistics()["average"]
        # σ = self.statistics()["stdev"]

        def normalise(a):
            return a
            # return (a - μ) / σ

        def generate():
            fields = []
            for s in self:
                fields.append(normalise(s.to_numpy()))
                if len(fields) >= offset:
                    yield fields[0], fields[-1]
                    fields.pop(0)

        import tensorflow as tf

        shape = self.first.shape

        dtype = kwargs.get("dtype", tf.float32)
        return tf.data.Dataset.from_generator(
            generate,
            output_signature=(
                tf.TensorSpec(shape, dtype=dtype, name="input"),
                tf.TensorSpec(shape, dtype=dtype, name="output"),
            ),
        )

    def _to_tfdataset_unsupervised(self, **kwargs):
        def generate():
            for s in self:
                yield s.to_numpy()

        import tensorflow as tf

        # TODO check the cost of the conversion
        # maybe default to float64
        dtype = kwargs.get("dtype", tf.float32)
        return tf.data.Dataset.from_generator(generate, dtype)

    def _to_tfdataset_supervised(self, label, **kwargs):
        @call_counter
        def generate():
            for s in self:
                yield s.to_numpy(), s.handle.get(label)

        import tensorflow as tf

        # with timer("_to_tfdataset_supervised shape"):
        shape = self.first.shape

        # TODO check the cost of the conversion
        # maybe default to float64
        dtype = kwargs.get("dtype", tf.float32)
        # with timer("tf.data.Dataset.from_generator"):
        return tf.data.Dataset.from_generator(
            generate,
            output_signature=(
                tf.TensorSpec(shape, dtype=dtype, name="data"),
                tf.TensorSpec(tuple(), dtype=tf.int64, name=label),
            ),
        )

    def to_xarray(self, **kwargs):

        import xarray as xr

        xarray_open_dataset_kwargs = {}

        if "xarray_open_mfdataset_kwargs" in kwargs:
            warnings.warn(
                "xarray_open_mfdataset_kwargs is deprecated, please use xarray_open_dataset_kwargs instead."
            )
            kwargs["xarray_open_dataset_kwargs"] = kwargs.pop(
                "xarray_open_mfdataset_kwargs"
            )

        user_xarray_open_dataset_kwargs = kwargs.get("xarray_open_dataset_kwargs", {})

        # until ignore_keys is included into cfgrib,
        # it is implemented here directly
        ignore_keys = user_xarray_open_dataset_kwargs.get("backend_kwargs", {}).pop(
            "ignore_keys", []
        )

        for key in ["backend_kwargs"]:
            xarray_open_dataset_kwargs[key] = mix_kwargs(
                user=user_xarray_open_dataset_kwargs.pop(key, {}),
                default={"errors": "raise"},
                forced={},
                logging_owner="xarray_open_dataset_kwargs",
                logging_main_key=key,
            )
        xarray_open_dataset_kwargs.update(
            mix_kwargs(
                user=user_xarray_open_dataset_kwargs,
                default={"squeeze": False},
                forced={
                    "errors": "raise",
                    "engine": "cfgrib",
                },
            )
        )

        result = xr.open_dataset(
            FieldsetAdapter(self, ignore_keys=ignore_keys),
            **xarray_open_dataset_kwargs,
        )

        def math_prod(lst):
            if not hasattr(math, "prod"):
                # python 3.7 does not have math.prod
                n = 1
                for x in lst:
                    n = n * x
                return n
            return math.prod(lst)

        def number_of_gribs(da):
            # Assumes last two dimensions are lat/lon coordinates
            skip = 2
            if da.dims[-1] == "values":
                # Assumes last dimension is the one-dimensional
                # lat/lon coordinate (non-regular grid)
                skip = 1
            return math_prod(list(da.shape)[:-skip])

        two_d_fields = sum(number_of_gribs(result[v]) for v in result.data_vars)

        # Make sure all the fields are converted
        # There may be more 2D xarray fields than GRB fields
        # if some missing dimension are filled with NaN values

        assert two_d_fields >= len(self), (
            "Not all GRIB fields were converted to xarray"
            f" ({len(self)} GRIBs > {two_d_fields} 2D-field(s) in xarray)"
        )

        return result

    def to_metview(self):
        from climetlab.metview import mv_read

        return mv_read(self.path)

    def to_numpy(self):
        import numpy as np

        return np.array([f.to_numpy() for f in self])

    def plot_map(self, backend):
        return self.first.plot_map(backend)

    def plot_graph(self, backend):
        import numpy as np

        what = backend._options("what", "global_average")
        what = dict(
            global_average=np.mean,
        )[what]

        # initialize list of lists
        data = [[s.valid_datetime(), what(s.to_numpy())] for s in self]
        import pandas as pd

        df = pd.DataFrame(data, columns=["date", "param"])

        backend.plot_graph_add_timeserie(df)

    # Used by normalisers
    def to_datetime(self):
        times = self.to_datetime_list()
        assert len(times) == 1
        return times[0]

    def to_datetime_list(self):
        # TODO: check if that can be done faster
        result = set()
        for s in self:
            result.add(s.valid_datetime())
        return sorted(result)

    def to_bounding_box(self):
        return BoundingBox.multi_merge([s.to_bounding_box() for s in self])

    def statistics(self):
        import numpy as np

        if self._statistics is not None:
            return self._statistics

        if False:
            cache = auxiliary_cache_file(
                "grib-statistics--",
                self.path,
                content="null",
                extension=".json",
            )

            with open(cache) as f:
                self._statistics = json.load(f)

            if self._statistics is not None:
                return self._statistics

        stdev = None
        average = None
        maximum = None
        minimum = None
        count = 0

        for s in self:
            v = s.values
            if count:
                stdev = np.add(stdev, np.multiply(v, v))
                average = np.add(average, v)
                maximum = np.maximum(maximum, v)
                minimum = np.minimum(minimum, v)
            else:
                stdev = np.multiply(v, v)
                average = v
                maximum = v
                minimum = v

            count += 1

        nans = np.count_nonzero(np.isnan(average))
        assert nans == 0, "Statistics with missing values not yet implemented"

        maximum = np.amax(maximum)
        minimum = np.amin(minimum)
        average = np.mean(average) / count
        stdev = np.sqrt(np.mean(stdev) / count - average * average)

        self._statistics = dict(
            minimum=minimum,
            maximum=maximum,
            average=average,
            stdev=stdev,
            count=count,
        )

        if False:
            with open(cache, "w") as f:
                json.dump(self._statistics, f)

        return self._statistics

    # def graph(self):
    #     # Compatibility with multi
    #     print("FieldSet.graph() not yet implemented")

    def save(self, filename):
        with open(filename, "wb") as f:
            self.write(f)

    def write(self, f):
        for s in self:
            s.write(f)
